<?php

namespace Webt\Model;

defined( '_JEXEC' ) or die;

use Joomla\CMS\Factory;
use Joomla\CMS\Log\Log;
use Joomla\Component\Fields\Administrator\Table\FieldTable;

/**
 * The FieldModel class is responsible for handling field-related database operations.
 *
 * @license     GNU General Public License version 2 or later, see License.txt
 */
class FieldModel extends BaseTranslatableModel {

	public $translatableFields = array(
		'title',
		'label',
		'default_value',
		'note',
		'description',
		'value',
	);

    public $translatableTypes = array(
        'text',
        'textarea',
        'editor'
    );
	
	/**
	 * Checks if field is translatable.
	 *
	 * @param object $field Field.
	 * @return boolean
	 */
	public static function isFieldTranslatable( $field ) {
		return '*' !== $field->language;
	}

	/**
	 * Checks if field value is translatable.
	 *
	 * @param object $field Field.
	 * @return boolean
	 */
	public function isFieldValueTranslatable( $field ) {
		return in_array( $field->type, $this->translatableTypes, true );
	}

	/**
	 * Retrieves all fields with values of an article.
	 *
	 * @param int $article_id Article ID.
	 * @return array An array of article objects.
	 */
	public function getArticleFields( int $article_id ) {
		$query = $this->_db->getQuery( true );

		$conditions = array(
			$this->_db->quoteName( 'item_id' ) . ' = ' . $article_id,
		);

		$query
			->select( 'f.*, fv.value, GROUP_CONCAT(value) as `values`' )
			->from( $this->_db->quoteName( '#__fields_values', 'fv' ) )
			->join(
				'INNER',
				$this->_db->quoteName( '#__fields', 'f' ) . ' ON ' .
				$this->_db->quoteName( 'f.id' ) . ' = ' . $this->_db->quoteName( 'fv.field_id' ) . ' AND ' .
				$this->_db->quoteName( 'f.context' ) . ' = ' . $this->_db->quote( 'com_content.article' )
			)
			->where( $conditions, 'AND' )
            ->group( $this->_db->quoteName( 'fv.field_id' ) . ',' . $this->_db->quoteName( 'fv.item_id' ) );

		$this->_db->setQuery( $query );

		return $this->_db->loadObjectList();
	}

	/**
	 * Get field translation in given language code.
	 *
	 * @param object $field Original field.
	 * @param string $target_language Target language code.
	 * @return object Field.
	 */
	public function getFieldTranslation( $field, $target_language ) {
		$query = $this->_db->getQuery( true );
		$name  = $field->name . "-$target_language";

		$conditions = array(
			$this->_db->quoteName( 'name' ) . ' = ' . $this->_db->quote( $name ),
			$this->_db->quoteName( 'language' ) . ' = ' . $this->_db->quote( $target_language ),
			$this->_db->quoteName( 'context' ) . ' = ' . $this->_db->quote( 'com_content.article' ),
		);

		$query
			->select( '*' )
			->from( $this->_db->quoteName( '#__fields' ) )
			->where( $conditions, 'AND' )
			->setLimit( 1 );

		$this->_db->setQuery( $query );

		return $this->_db->loadObject();
	}

	/**
	 * Get value by field and article IDs.
	 *
	 * @param int $article_id Article ID.
	 * @param int $field_id Field ID.
	 * @return object Field value object.
	 */
	public function getFieldValue( $article_id, $field_id ) {
		$query = $this->_db->getQuery( true );

		$conditions = array(
			$this->_db->quoteName( 'field_id' ) . ' = ' . $field_id,
			$this->_db->quoteName( 'item_id' ) . ' = ' . $article_id,
		);

		$query
			->select( '*' )
			->from( $this->_db->quoteName( '#__fields_values' ) )
			->where( $conditions, 'AND' );

		$this->_db->setQuery( $query );

		return $this->_db->loadObject();
	}

	/**
	 * Update existing field value.
	 *
	 * @param int    $article_id Article ID.
	 * @param int    $field_id   Field ID.
	 * @param string $value New value.
	 * @return void
	 */
	public function updateFieldValue( $article_id, $field_id, $value ) {
		Log::add( "Updating article $article_id field $field_id value", Log::DEBUG, 'webt' );

		$query = $this->_db->getQuery( true );

		$conditions = array(
			$this->_db->quoteName( 'field_id' ) . ' = ' . $field_id,
			$this->_db->quoteName( 'item_id' ) . ' = ' . $article_id,
		);

		$query
			->update( '#__fields_values' )
			->set( $this->_db->quoteName( 'value' ) . ' = ' . $this->_db->quote( $value ) )
			->where( $conditions );

		$this->_db->setQuery( $query );
		$this->_db->execute();
		Log::add( "Updated article $article_id field $field_id value", Log::DEBUG, 'webt' );
	}

	/**
	 * Inserts field value in database.
	 *
	 * @param int $article_id Article ID.
	 * @param int $field_id   Field ID.
	 * @param int $value      Field value.
	 * @return array An array of article objects.
	 */
	public function insertFieldValue( $article_id, $field_id, $value ) {
		Log::add( "Inserting field $field_id value for article $article_id", Log::DEBUG, 'webt' );

		$existingFieldValue = $this->getFieldValue( $article_id, $field_id );
		if ( $existingFieldValue ) {
			$this->updateFieldValue( $article_id, $field_id, $value );
			return;
		} else {
			$field_value           = new \stdClass();
			$field_value->field_id = $field_id;
			$field_value->item_id  = $article_id;
			$field_value->value    = $value;

			$this->_db->insertObject( '#__fields_values', $field_value );
		}
		Log::add( "Inserted field $field_id value for article $article_id", Log::DEBUG, 'webt' );
	}

	/**
	 * Insert translated version of source field
	 *
	 * @param object   $source_field Source field.
	 * @param string[] $translations Translations.
	 * @param string   $language Target language code.
	 * @return int Field ID.
	 */
	public function insertFieldTranslation( $source_field, $translations, $language ) {
		Log::add( "Inserting field $source_field->id translation", Log::DEBUG, 'webt' );

		$date     = Factory::getDate()->toSql();

		$field_table   = new FieldTable( $this->_db );
		$default_value = self::isFieldValueTranslatable( $source_field ) ? $translations['default_value'] : $source_field->default_value;

		$field_table->title               = $translations['title'];
		$field_table->label               = $translations['label'];
		$field_table->default_value       = $default_value;
		$field_table->note                = $translations['note'];
		$field_table->description         = $translations['description'];
		$field_table->name                = $source_field->name . "-$language";
		$field_table->context             = $source_field->context;
		$field_table->type                = $source_field->type;
		$field_table->state               = $source_field->state;
		$field_table->only_use_in_subform = $source_field->only_use_in_subform;
		$field_table->checked_out         = $source_field->checked_out;
		$field_table->checked_out_time    = $source_field->checked_out_time;
		$field_table->ordering            = $source_field->ordering;
		$field_table->params              = $source_field->params;
		$field_table->fieldparams         = $source_field->fieldparams;
		$field_table->group_id            = $source_field->group_id;
		$field_table->created_user_id     = $source_field->created_user_id;
		$field_table->modified_by         = $source_field->modified_by;
		$field_table->access              = $source_field->access;
		$field_table->language            = $language;
		$field_table->created_time        = $date;
		$field_table->modified_time       = $date;
		$field_table->required            = 0;

        $field_table->store();

		Log::add( "Inserted field $field_table->id", Log::DEBUG, 'webt' );
		return $field_table->id;
	}
}
