<?php

defined( '_JEXEC' ) or die;

use Joomla\CMS\Date\Date;
use Joomla\CMS\Factory;
use Joomla\CMS\Log\Log;
use Joomla\CMS\Plugin\CMSPlugin;
use Joomla\CMS\Uri\Uri;
use Joomla\CMS\Language\Text;

JLoader::registerNamespace( 'Webt', JPATH_LIBRARIES . '/Webt' );

use Webt\Model\SettingsModel;
use Webt\Model\AssociationModel;
use Webt\Model\ArticleModel;
use Webt\Model\EtranslationResponseModel;
use Webt\Model\FieldModel;
use Webt\Services\EtranslationUtils;
use Webt\Services\TranslationManager;
/**
 * @copyright   Copyright (C) 2005 - 2008 Open Source Matters. All rights reserved.
 * @license     GNU General Public License version 2 or later, see License.txt
 */

// Check to ensure this file is included in Joomla!
defined( '_JEXEC' ) or die( 'Restricted access' );


/**
 * Joomla plugin for managing article translations using the Webt component.
 */
class plgContentWebt extends CMSPlugin {

	/**
	 * Load the language file on instantiation.
	 *
	 * @var    boolean
	 * @since  3.1
	 */
	protected $autoloadLanguage = true;

	// Message keys
	private $schedule_association_creation_key = 'webt_scheduled_associations';
	private $article_changed_key               = 'webt_article_changed';
	/**
	 * Original text used in MT notice. Used only to check if it has been overriden, actual text displayed is retrieved from ini file.
	 *
	 * @var string
	 */
	private $original_notice_text = 'This page has been machine-translated.';
	/**
	 * Stores IDs of articles checked during page load.
	 *
	 * @var array
	 */
	private $articles_checked = array();

	/**
	 * Method that happens before content is saved
	 * Checks if there is a new translation for an article
	 *
	 * Method is called right before content is saved into the database.
	 * Article object is passed by reference, so any changes will be saved!
	 * NOTE:  Returning false will abort the save with an error.
	 *  You can set the error by calling $article->setError($message)
	 *
	 * @param   object      A JTableContent object
	 * @param   bool        If the content is just about to be created
	 * @return  bool        If false, abort the save
	 */
	function onContentBeforeSave( $context, &$article, $isNew, $data ) {
		$settingsModel = new SettingsModel();
		if ( $settingsModel->getTranslateOnSave() ) {
			$isArticle = ( 'com_content.article' === $context );
			$isSource  = ( $article->language === $settingsModel->getDefaultLanguage() );
			$isTrashed = ( -2 === $article->state );

			if ( $isArticle && $isSource && ! $isNew && ! $isTrashed && $this->articleContentChanged( $article, $data ) ) {
				// Mark article as changed.
				$session = Factory::getSession();
				$session->set( $this->article_changed_key, $article->id );
				Log::add( "Article $article->id has been changed and is marked for retranslation.", Log::DEBUG, 'webt' );
			}
		}
		return true;
	}

	/**
	 * Checks if article has changed upon saving article.
	 *
	 * @param object $article Updated article.
	 * @param array  $submitted_data Submitted data.
	 * @return bool
	 */
	private function articleContentChanged( $article, $submitted_data ) {
		$articleModel       = new ArticleModel();
		$translationManager = new TranslationManager();

		// Check article properties.
		$previousVersion = $articleModel->getArticleById( $article->id );
		$previousValues  = $translationManager->getTranslatableStrings( $previousVersion );
		$newValues       = $translationManager->getTranslatableStrings( $article );

		if ( $newValues != $previousValues ) {
			return true;
		}

		// Check fields.
		return $this->articleFieldsChanged( $article, $submitted_data );
	}

	/**
	 * Checks if article fields has changed upon saving article.
	 *
	 * @param object $article Updated article.
	 * @param array  $submitted_data Submitted data.
	 * @return bool
	 */
	private function articleFieldsChanged( $article, $submitted_data ) {
		$fieldModel = new FieldModel();
		$fields     = $fieldModel->getArticleFields( $article->id );

		$submitted_field_names = array_keys( $submitted_data['com_fields'] );
		$existing_field_names  = array_map(
			function( $field ) {
				return $field->name;
			},
			$fields
		);
		$field_names           = array_unique( array_merge( $submitted_field_names, $existing_field_names ) );

		// Check each field value.
		foreach ( $field_names as $field_name ) {
			$field = current(
				array_filter(
					$fields,
					function( $f ) use ( $field_name ) {
						return $f->name === $field_name;
					}
				)
			);
			if ( ! $field && ! $submitted_data['com_fields'][ $field_name ] ) {
				// Both values are empty, skip.
				continue;
			}

			$oldValue = $field->value;
			$newValue = $submitted_data['com_fields'][ $field_name ];

			// Handle special field types.
			if ( 'calendar' === $field->type ) {
				// Ignore time part which is updated to submission time.
				$oldValue = $oldValue ? ( new Date( $oldValue ) )->format( 'Y-m-d' ) : null;
				$newValue = $newValue ? ( new Date( $newValue ) )->format( 'Y-m-d' ) : null;
			} elseif ( 'checkboxes' === $field->type ) {
				// Use grouped value, as checkbox values are stored separately.
				$oldValue = explode( ',', $field->values );
			} elseif ( in_array( $field->type, array( 'media', 'subform' ), true ) ) {
				// Decode JSON value as array to match form data.
				$oldValue = json_decode( $oldValue, true );
			}

			if ( $oldValue != $newValue ) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Event handler for the `onContentAfterSave` event.
	 * Called after the content is saved into the database.
	 *
	 * @param string $context Component context.
	 * @param object $article A reference to the saved article.
	 * @param bool   $isNew True if the article is newly created.
	 * @param array  $data Submitted form data.
	 * @return bool Return true to continue processing or false to halt.
	 */
	function onContentAfterSave( $context, &$article, $isNew, $data ) {
		$settingsModel = new SettingsModel();
		if ( $settingsModel->getTranslateOnSave() ) {
			$sourceLanguage = $settingsModel->getDefaultLanguage();

			if ( $context == 'com_content.article' && $article->language == $sourceLanguage && $article->state != -2 ) {
				$force_retranslate = true;

				$session = Factory::getSession();
				if ( ! $isNew ) {
					$changed_article_id = $session->get( $this->article_changed_key );
					$session->clear( $this->article_changed_key );
					if ( ! $changed_article_id || $changed_article_id !== $article->id ) {
						// Article translatable content not changed, do not force retranslation.
						Log::add( "Article [ID = $article->id] content has not changed, skipping force retranslation.", Log::DEBUG, 'webt' );
						$force_retranslate = false;
					}
				}

				$targetLanguageArray = array_filter(
					$this->getMissingTranslationLangCodes( $article, $force_retranslate ),
					function( $langCode ) use ( $isNew, $data, $force_retranslate ) {
						return ( ! $isNew && $force_retranslate ) || ! $data['associations'][ $langCode ]; // allow manually selected associations.
					}
				);

				if ( ! empty( $targetLanguageArray ) ) {
					Log::add( "Translating updated source article on save [ID=$article->id]", Log::DEBUG, 'webt' );
					$translationManager = new TranslationManager();
					$new_articles       = $translationManager->translateArticle( $article, $targetLanguageArray, true, false );

					// Joomla recreates submitted article associations during save event, so we need to schedule association creation in a later event.
					$session->set( $this->schedule_association_creation_key, $new_articles );

					Factory::getApplication()->enqueueMessage( 'Article translated!' );
				}
			}
		}
		return true;
	}

	/**
	 * Listener for the `onAfterDispatch` event
	 *
	 * @return  void
	 *
	 * @since   1.0
	 */
	public function onAfterDispatch() {
		// Create translated article associations.
		$session      = Factory::getSession();
		$new_articles = $session->get( $this->schedule_association_creation_key );

		if ( $new_articles !== null ) {
			$associationModel = new AssociationModel();
			foreach ( $new_articles as $source_article_id => $translation_ids ) {
				$existing_association = $associationModel->getArticleAssociation( $source_article_id );

				if ( $existing_association ) {
					$key = $existing_association->key;
				} else {
					$key = md5( json_encode( $new_articles ) );
					$associationModel->createArticleAssociation( $source_article_id, $key );
				}

				foreach ( $translation_ids as $articleId ) {
					$associationModel->createArticleAssociation( $articleId, $key );
				}
			}
			$session->clear( $this->schedule_association_creation_key );
		}
	}

	/**
	 * Listener for the `onContentPrepare` event.
	 *
	 * @param string  $context Component name and view or name of module.
	 * @param mixed   $article A reference to the article that is being rendered by the view.
	 * @param mixed   $params A reference to an associative array of relevant parameters.
	 * @param integer $page An integer that determines the "page" of the content that is to be generated.
	 * @return void
	 */
	public function onContentPrepare( $context, &$article, &$params, $page ) {
		if ( Factory::getApplication()->isClient( 'site' ) && 'com_content.article' === $context ) {
			$this->addMtNotice();
			$this->checkTranslations( $article );
		}
	}

	/**
	 * Adds MT notice disclaimer.
	 *
	 * @return void
	 */
	private function addMtNotice() {
		$text          = Text::_( 'MT_NOTICE_TEXT' );
		$settingsModel = new SettingsModel();

		if ( $text && $settingsModel->getShowMtNotice() ) {
			$sourceLanguage  = $settingsModel->getDefaultLanguage();
			$currentLanguage = Factory::getLanguage()->getTag();

			if ( $sourceLanguage !== $currentLanguage ) {
				$isNoticeTranslated = $text !== $this->original_notice_text;
				$document           = Factory::getDocument();
				$asset_dir          = Uri::root() . 'administrator/components/com_webt/assets';
				$document->addScriptOptions(
					'webt',
					array(
						'mt_notice_text' => $text,
						'mt_notice_lang' => $isNoticeTranslated ? $currentLanguage : $sourceLanguage,
					)
				);
				$document->addScript( "$asset_dir/js/mt_notice.js" );
				$document->addStyleSheet( "$asset_dir/css/mt_notice.css" );
			}
		}
	}

	/**
	 * Checks if article has all translations and translates to missing languages
	 *
	 * @param mixed $article Article to be loaded.
	 * @return void
	 */
	private function checkTranslations( $article ) {
		if ( in_array( $article->id, $this->articles_checked ) ) {
			return;
		}

		Log::add( "Checking if article $article->id has translations...", Log::DEBUG, 'webt' );
		$settingsModel = new SettingsModel();
		if ( $settingsModel->getTranslateOnVisit() ) {
			$articleModel   = new ArticleModel();
			$sourceLanguage = $settingsModel->getDefaultLanguage();
			$sourceArticle  = $article;
			if ( $article->language !== $sourceLanguage ) {
				$sourceArticle = $articleModel->getArticleTranslation( $article->id, $sourceLanguage );
			}
			if ( $sourceArticle ) {
				if ( in_array( $sourceArticle->id, $this->articles_checked ) ) {
					return;
				}
				$this->articles_checked = array_merge( $this->articles_checked, array( $article->id, $sourceArticle->id ) );

				$missingTranslations = array_filter(
					$this->getMissingTranslationLangCodes( $sourceArticle ),
					function( $langCode ) use ( &$sourceArticle ) {
						return ! self::checkLateTranslation( $sourceArticle, $langCode );
					}
				);

				if ( ! empty( $missingTranslations ) ) {
					$stringLanguages = implode( ', ', $missingTranslations );
					Log::add( "Translating article on page load to missing languages ($stringLanguages) [ID=$sourceArticle->id]", Log::INFO, 'webt' );
					$translationManager = new TranslationManager();
					$translationManager->translateArticle( $sourceArticle, $missingTranslations, false, true, false );
				}
			}
		}
	}

	/**
	 * Gets translation language codes for article
	 *
	 * @param object  $article Article.
	 * @param boolean $force_retranslate If true, returns all translation language codes.
	 * @return string[] Language code list.
	 */
	private function getMissingTranslationLangCodes( $article, $force_retranslate = false ) {
		$settingsModel       = new SettingsModel();
		$targetLanguageArray = $settingsModel->getLanguages();
		$languageCodeArray   = array_column( $targetLanguageArray, 'lang_code' );

		if ( $force_retranslate ) {
			return $languageCodeArray;
		}

		return array_filter(
			$languageCodeArray,
			function( $langCode ) use ( &$article ) {
				$articleModel = new ArticleModel();
				return ! $articleModel->getArticleTranslation( $article->id, $langCode );
			}
		);
	}

	/**
	 * Checks if there is a late translation and applies it to the article.
	 *
	 * @param int    $article_id The ID of the article to check for late translation.
	 * @param string $langcode The language code to check for late translation.
	 * @return bool Returns true if a late translation was applied, false otherwise.
	 */
	private static function checkLateTranslation( $article, $langcode ) {
		$settingsModel = new SettingsModel();
		if ( 'etranslation' === $settingsModel->getMtEngine() ) {
			$responseModel = new EtranslationResponseModel();
			$entry         = $responseModel->getTranslationEntryByArticleIdAndLangcode( $article->id, $langcode );
			if ( ! $entry ) {
				return false;
			}
			$is_translation = ! in_array( $entry->response, array( EtranslationUtils::$initial_value, EtranslationUtils::$timeout_value, EtranslationUtils::$error_value ), true );
			if ( $entry && $is_translation ) {
				Log::add( "Found late eTranslation response entry for article [ID=$article->id]", Log::INFO, 'webt' );
				$responseModel->deleteTranslationEntry( $responseModel->request_id );

				$translations       = EtranslationUtils::decode_response( $entry->response );
				$translationManager = new TranslationManager();
				$translationManager->translateArticle( $article, array( $langcode ), true, true, false, $translations );
				return true;
			}
		}
		return false;
	}
}
